@@ -517,7 +517,9 @@ ical_attendees_from_props (icalcomponent *ical_comp, GSList *recipients, gboolea
 			prop = icalproperty_new_organizer (val);
 
 			/* CN */
-			str = (const gchar *) exchange_mapi_util_find_row_propval (&recip->out_SRow, PR_RECIPIENT_DISPLAY_NAME_UNICODE);
+			str = recip->display_name;
+			if (!str || !*str)
+				str = (const gchar *) exchange_mapi_util_find_row_propval (&recip->out_SRow, PR_RECIPIENT_DISPLAY_NAME_UNICODE);
 			if (!str)
 				str = (const gchar *) exchange_mapi_util_find_row_propval (&recip->out_SRow, PR_DISPLAY_NAME_UNICODE);
 			if (str) {
@@ -528,7 +530,9 @@ ical_attendees_from_props (icalcomponent *ical_comp, GSList *recipients, gboolea
 			prop = icalproperty_new_attendee (val);
 
 			/* CN */
-			str = (const gchar *) exchange_mapi_util_find_row_propval (&recip->out_SRow, PR_RECIPIENT_DISPLAY_NAME_UNICODE);
+			str = recip->display_name;
+			if (!str || !*str)
+				str = (const gchar *) exchange_mapi_util_find_row_propval (&recip->out_SRow, PR_RECIPIENT_DISPLAY_NAME_UNICODE);
 			if (!str)
 				str = (const gchar *) exchange_mapi_util_find_row_propval (&recip->out_SRow, PR_DISPLAY_NAME_UNICODE);
 			if (str) {
@@ -568,11 +572,14 @@ static const uint8_t GID_START_SEQ[] = {
 	0x74, 0xc5, 0xb7, 0x10, 0x1a, 0x82, 0xe0, 0x08
 };
 
+/* exception_replace_time is a value of PidLidExceptionReplaceTime; this is not used for 'clean' object ids.
+   creation_time is a value of PR_CREATION_TIME
+*/
 void
-exchange_mapi_cal_util_generate_globalobjectid (gboolean is_clean, const gchar *uid, struct Binary_r *sb)
+exchange_mapi_cal_util_generate_globalobjectid (gboolean is_clean, const gchar *uid, const struct timeval *exception_replace_time, const struct FILETIME *creation_time, struct Binary_r *sb)
 {
 	GByteArray *ba;
-	guint32 flag32;
+	guint32 val32;
 	guchar *buf = NULL;
 	gsize len;
 	d(guint32 i);
@@ -581,36 +588,40 @@ exchange_mapi_cal_util_generate_globalobjectid (gboolean is_clean, const gchar *
 
 	ba = g_byte_array_append (ba, GID_START_SEQ, (sizeof (GID_START_SEQ) / sizeof (GID_START_SEQ[0])));
 
-	/* FIXME for exceptions */
-	if (is_clean || TRUE) {
-		flag32 = 0;
-		ba = g_byte_array_append (ba, (const guint8 *)&flag32, sizeof (guint32));
+	val32 = 0;
+	if (!is_clean && exception_replace_time) {
+		struct icaltimetype icaltm = icaltime_from_timet_with_zone (exception_replace_time->tv_sec, 0, icaltimezone_get_utc_timezone ());
+
+		val32 |= (icaltm.year & 0xFF00) << 16;
+		val32 |= (icaltm.year & 0xFF) << 16;
+		val32 |= (icaltm.month & 0xFF) << 8;
+		val32 |= (icaltm.day & 0xFF);
 	}
 
-	/* creation time - may be all 0's  */
-	flag32 = 0;
-	ba = g_byte_array_append (ba, (const guint8 *)&flag32, sizeof (guint32));
-	flag32 = 0;
-	ba = g_byte_array_append (ba, (const guint8 *)&flag32, sizeof (guint32));
+	ba = g_byte_array_append (ba, (const guint8 *) &val32, sizeof (guint32));
+
+	/* creation time */
+	val32 = creation_time ? creation_time->dwLowDateTime : 0;
+	ba = g_byte_array_append (ba, (const guint8 *) &val32, sizeof (guint32));
+	val32 = creation_time ? creation_time->dwHighDateTime : 0;
+	ba = g_byte_array_append (ba, (const guint8 *) &val32, sizeof (guint32));
 
 	/* RESERVED - should be all 0's  */
-	flag32 = 0;
-	ba = g_byte_array_append (ba, (const guint8 *)&flag32, sizeof (guint32));
-	flag32 = 0;
-	ba = g_byte_array_append (ba, (const guint8 *)&flag32, sizeof (guint32));
-
-	/* FIXME: cleanup the UID first */
+	val32 = 0;
+	ba = g_byte_array_append (ba, (const guint8 *) &val32, sizeof (guint32));
+	val32 = 0;
+	ba = g_byte_array_append (ba, (const guint8 *) &val32, sizeof (guint32));
 
 	/* We put Evolution's UID in base64 here */
 	buf = g_base64_decode (uid, &len);
 	if (len % 2 != 0)
 		--len;
-	flag32 = len;
+	val32 = len;
 
 	/* Size in bytes of the following data */
-	ba = g_byte_array_append (ba, (const guint8 *)&flag32, sizeof (guint32));
+	ba = g_byte_array_append (ba, (const guint8 *) &val32, sizeof (guint32));
 	/* Data */
-	ba = g_byte_array_append (ba, (const guint8 *)buf, flag32);
+	ba = g_byte_array_append (ba, (const guint8 *)buf, val32);
 	g_free (buf);
 
 	sb->lpb = ba->data;
@@ -623,17 +634,18 @@ exchange_mapi_cal_util_generate_globalobjectid (gboolean is_clean, const gchar *
 	g_byte_array_free (ba, FALSE);
 }
 
+/* returns complete globalid as base64 encoded string */
 static gchar *
-id_to_string (GByteArray *ba)
+globalid_to_string (GByteArray *ba)
 {
 	guint8 *ptr;
 	guint len;
-	gchar *buf = NULL;
-	guint32 flag32, i, j;
+	guint32 i, j;
 
 	g_return_val_if_fail (ba != NULL, NULL);
+
 	/* MSDN docs: the globalID must have an even number of bytes */
-	if ((ba->len)%2 != 0)
+	if ((ba->len) % 2 != 0)
 		return NULL;
 
 	ptr = ba->data;
@@ -644,53 +656,12 @@ id_to_string (GByteArray *ba)
 		if (*ptr != GID_START_SEQ[j])
 			return NULL;
 
-	/* FIXME: for exceptions - len = 4 bytes */
-	flag32 = *((guint32 *)ptr);
-	i += sizeof (guint32);
-	if (!(i < len) || flag32 != 0)
-		return NULL;
-	ptr += sizeof (guint32);
-
-	/* Creation time - len = 8 bytes - skip it */
-	flag32 = *((guint32 *)ptr);
-	i += sizeof (guint32);
-	if (!(i < len))
-		return NULL;
-	ptr += sizeof (guint32);
-
-	flag32 = *((guint32 *)ptr);
-	i += sizeof (guint32);
-	if (!(i < len))
-		return NULL;
-	ptr += sizeof (guint32);
-
-	/* Reserved bytes - len = 8 bytes */
-	flag32 = *((guint32 *)ptr);
-	i += sizeof (guint32);
-	if (!(i < len) || flag32 != 0)
-		return NULL;
-	ptr += sizeof (guint32);
-
-	flag32 = *((guint32 *)ptr);
-	i += sizeof (guint32);
-	if (!(i < len) || flag32 != 0)
-		return NULL;
-	ptr += sizeof (guint32);
-
-	/* This is the real data */
-	flag32 = *((guint32 *)ptr);
-	i += sizeof (guint32);
-	if (!(i < len) || flag32 != (len - i))
-		return NULL;
-	ptr += sizeof (guint32);
-
-	buf = g_base64_encode (ptr, flag32);
-
-	return buf;
+	/* take complete global id */
+	return g_base64_encode (ba->data, ba->len);
 }
 
 ECalComponent *
-exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcomponent_kind kind, const gchar *mid, struct mapi_SPropValue_array *properties,
+exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, mapi_id_t fid, icalcomponent_kind kind, const gchar *mid, struct mapi_SPropValue_array *properties,
 					   GSList *streams, GSList *recipients, GSList *attachments,
 					   const gchar *local_store_uri, const icaltimezone *default_zone, gboolean is_reply, GSList **detached_components)
 {
@@ -698,6 +669,7 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 	struct timeval t;
 	ExchangeMAPIStream *body_stream;
 	const gchar *subject = NULL, *body = NULL;
+	const struct StringArray_r *categories_array;
 	const uint32_t *ui32;
 	const bool *b;
 	icalcomponent *ical_comp;
@@ -752,26 +724,35 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 	icalcomponent_set_summary (ical_comp, subject);
 	icalcomponent_set_description (ical_comp, body);
 
+	categories_array = exchange_mapi_util_find_array_namedid (properties, conn, fid, PidNameKeywords);
+	if (categories_array) {
+		GSList *categories = NULL;
+		gint ii;
+
+		for (ii = 0; ii < categories_array->cValues; ii++) {
+			const gchar *category = categories_array->lppszA[ii];
+
+			if (!category || !*category)
+				continue;
+
+			categories = g_slist_append (categories, (gpointer) category);
+		}
+
+		e_cal_component_set_categories_list (comp, categories);
+
+		g_slist_free (categories);
+	}
+
 	if (icalcomponent_isa (ical_comp) == ICAL_VEVENT_COMPONENT) {
 		const gchar *location = NULL;
 		const gchar *dtstart_tz_location = NULL, *dtend_tz_location = NULL;
 		gboolean all_day;
 		ExchangeMAPIStream *stream;
 
-		/* CleanGlobalObjectId */
-		stream = exchange_mapi_util_find_stream (streams, PidLidCleanGlobalObjectId);
-		if (stream) {
-			gchar *value = id_to_string (stream->value);
-			prop = icalproperty_new_x (value);
-			icalproperty_set_x_name (prop, "X-EVOLUTION-MAPI-CLEAN-GLOBALID");
-			icalcomponent_add_property (ical_comp, prop);
-			g_free (value);
-		}
-
 		/* GlobalObjectId */
 		stream = exchange_mapi_util_find_stream (streams, PidLidGlobalObjectId);
 		if (stream) {
-			gchar *value = id_to_string (stream->value);
+			gchar *value = globalid_to_string (stream->value);
 			prop = icalproperty_new_x (value);
 			icalproperty_set_x_name (prop, "X-EVOLUTION-MAPI-GLOBALID");
 			icalcomponent_add_property (ical_comp, prop);
@@ -788,6 +769,16 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 			g_free (value);
 		}
 
+		ui32 = find_mapi_SPropValue_data(properties, PR_OWNER_APPT_ID);
+		if (ui32) {
+			gchar *value = exchange_mapi_util_mapi_id_to_string ((mapi_id_t) (*ui32));
+
+			prop = icalproperty_new_x (value);
+			icalproperty_set_x_name (prop, "X-EVOLUTION-MAPI-OWNER-APPT-ID");
+			icalcomponent_add_property (ical_comp, prop);
+			g_free (value);
+		}
+
 		/* AppointmentSequence */
 		ui32 = (const uint32_t *)find_mapi_SPropValue_data(properties, PidLidAppointmentSequence);
 		if (ui32) {
@@ -830,7 +821,12 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 		}
 
 		if (get_mapi_SPropValue_array_date_timeval (&t, properties, PidLidAppointmentEndWhole) == MAPI_E_SUCCESS) {
-			icaltimezone *zone = dtend_tz_location ? icaltimezone_get_builtin_timezone (dtend_tz_location) : (icaltimezone *)default_zone;
+			icaltimezone *zone;
+
+			if (!dtend_tz_location)
+				dtend_tz_location = dtstart_tz_location;
+
+			zone = dtend_tz_location ? icaltimezone_get_builtin_timezone (dtend_tz_location) : (icaltimezone *) default_zone;
 			prop = icalproperty_new_dtend (icaltime_from_timet_with_zone (t.tv_sec, all_day, zone));
 			if (!all_day && zone && icaltimezone_get_tzid (zone)) {
 				icalproperty_add_parameter (prop, icalparameter_new_tzid (icaltimezone_get_tzid (zone)));
@@ -863,7 +859,7 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 						email = "";
 
 					if (g_str_equal (email_type, "EX")) {
-						to_free = exchange_mapi_connection_ex_to_smtp (conn, email, NULL);
+						to_free = exchange_mapi_connection_ex_to_smtp (conn, email, NULL, NULL);
 						email = to_free;
 					}
 
@@ -895,7 +891,7 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 						email = "";
 
 					if (g_str_equal (email_type, "EX")) {
-						to_free = exchange_mapi_connection_ex_to_smtp (conn, email, NULL);
+						to_free = exchange_mapi_connection_ex_to_smtp (conn, email, NULL, NULL);
 						email = to_free;
 					}
 
@@ -924,11 +920,11 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 				const gchar *sent_email = (const gchar *) exchange_mapi_util_find_array_propval (properties, PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE);
 
 				if (!g_utf8_collate (sender_email_type, "EX")) {
-					sender_free = exchange_mapi_connection_ex_to_smtp (conn, sender_email, NULL);
+					sender_free = exchange_mapi_connection_ex_to_smtp (conn, sender_email, NULL, NULL);
 					sender_email = sender_free;
 				}
 				if (!g_utf8_collate (sent_email_type, "EX")) {
-					sent_free = exchange_mapi_connection_ex_to_smtp (conn, sent_email, NULL);
+					sent_free = exchange_mapi_connection_ex_to_smtp (conn, sent_email, NULL, NULL);
 					sent_email = sent_free;
 				}
 
@@ -957,7 +953,15 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 		if (b && *b) {
 			stream = exchange_mapi_util_find_stream (streams, PidLidAppointmentRecur);
 			if (stream) {
-				exchange_mapi_cal_util_bin_to_rrule (stream->value, comp, detached_components);
+				icaltimezone *recur_zone;
+				const gchar *recur_tz_location;
+
+				recur_tz_location = exchange_mapi_util_find_array_namedid (properties, conn, fid, PidLidTimeZoneDescription);
+				if (recur_tz_location)
+					recur_tz_location = exchange_mapi_cal_tz_util_get_ical_equivalent (recur_tz_location);
+				recur_zone = recur_tz_location ? icaltimezone_get_builtin_timezone (recur_tz_location) : (icaltimezone *) default_zone;
+
+				exchange_mapi_cal_util_bin_to_rrule (stream->value, comp, detached_components, recur_zone);
 			}
 		}
 
@@ -1054,8 +1058,6 @@ exchange_mapi_cal_util_mapi_props_to_comp (ExchangeMapiConnection *conn, icalcom
 		icalcomponent_add_property (ical_comp, prop);
 	}
 
-	/* FIXME: categories */
-
 	set_attachments_to_cal_component (comp, attachments, local_store_uri);
 
 	e_cal_component_rescan (comp);
@@ -1092,7 +1094,7 @@ fetch_camel_cal_comp_cb (FetchItemsCallbackData *item_data, gpointer data)
 			smid = exchange_mapi_util_mapi_id_to_string (item_data->mid);
 		else
 			smid = e_cal_component_gen_uid();
-		comp = exchange_mapi_cal_util_mapi_props_to_comp (item_data->conn, fccd->kind, smid,
+		comp = exchange_mapi_cal_util_mapi_props_to_comp (item_data->conn, item_data->fid, fccd->kind, smid,
 							item_data->properties, item_data->streams, item_data->recipients,
 							item_data->attachments, filepath, NULL, fccd->method == ICAL_METHOD_REPLY,
 							&detached_recurrences);
@@ -1181,7 +1183,9 @@ exchange_mapi_cal_utils_add_named_ids (ExchangeMapiConnection *conn, mapi_id_t f
 		{ PidLidTaskMode, 0 },
 		{ PidLidReminderSignalTime, 0 },
 		{ PidLidTimeZoneStruct, 0 },
-		{ PidLidTimeZoneDescription, 0 }
+		{ PidLidTimeZoneDescription, 0 },
+		{ PidLidExceptionReplaceTime, 0 },
+		{ PidNameKeywords, 0 }
 	};
 	icalcomponent_kind kind = pkind;
 
@@ -1338,6 +1342,7 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 	const gchar *dtstart_tz_location, *dtend_tz_location, *text = NULL;
 	time_t tt;
 	gboolean is_all_day;
+	GSList *categories = NULL;
 
 	g_return_val_if_fail (conn != NULL, FALSE);
 	g_return_val_if_fail (mem_ctx != NULL, FALSE);
@@ -1460,6 +1465,33 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 	set_value (PR_BODY_UNICODE, text);
 	text = NULL;
 
+	e_cal_component_get_categories_list (comp, &categories);
+	if (categories) {
+		gint ii;
+		GSList *c;
+		struct StringArray_r *categories_array;
+
+		categories_array = talloc_zero (mem_ctx, struct StringArray_r);
+		categories_array->cValues = g_slist_length (categories);
+		categories_array->lppszA = (const char **) talloc_zero_array (mem_ctx, gchar *, categories_array->cValues);
+
+		for (c = categories, ii = 0; c; c = c->next, ii++) {
+			const gchar *category = c->data;
+
+			if (!category || !*category) {
+				ii--;
+				categories_array->cValues--;
+				continue;
+			}
+
+			categories_array->lppszA[ii] = talloc_strdup (mem_ctx, category);
+		}
+
+		set_named_value (PidNameKeywords, categories_array);
+
+		e_cal_component_free_categories_list (categories);
+	}
+
 	/* Priority and Importance */
 	prop = icalcomponent_get_first_property (ical_comp, ICAL_PRIORITY_PROPERTY);
 	flag32 = prop ? get_prio_prop_from_priority (icalproperty_get_priority (prop)) : PRIORITY_NORMAL;
@@ -1649,6 +1681,19 @@ exchange_mapi_cal_utils_write_props_cb (ExchangeMapiConnection *conn, mapi_id_t
 		set_named_value (PidLidRecurrenceType, &flag32);
 
 		flag32 = cbdata->appt_id;
+		if (!flag32) {
+			gchar *propval;
+
+			propval = exchange_mapi_cal_utils_get_icomp_x_prop (e_cal_component_get_icalcomponent (comp), "X-EVOLUTION-MAPI-OWNER-APPT-ID");
+			if (propval && *propval) {
+				mapi_id_t as_id = 0;
+
+				if (exchange_mapi_util_mapi_id_from_string (propval, &as_id))
+					flag32 = (uint32_t) as_id;
+			}
+
+			g_free (propval);
+		}
 		set_value (PR_OWNER_APPT_ID, &flag32);
 
 		flag32 = cbdata->appt_seq;
